class InstanceNotFound(Exception):
    def __init__(self):
        super().__init__("The specified instance url does not exist")


class IncorrectCredentials(Exception):
    def __init__(self):
        super().__init__("Incorrect email or password")



import gzip
import shutil


class ObjectS3:
    def __init__(self, path):
        self.__path = path

    @property
    def path(self):
        return self.__path

    def is_compressed(self):
        with gzip.open(self.__path, "r") as fh:
            try:
                # Try to read 1 byte of the file (Fails if not zipped)
                fh.read1()
                return True
            except Exception:
                return False

    def compress(self):
        with open(self.__path, "rb") as f_in:
            compressed_path = self.__path + ".gz"
            with gzip.open(compressed_path, "wb") as f_out:
                # Copying is done in chunks by default
                shutil.copyfileobj(f_in, f_out)
                self.__path = compressed_path



import json
from importlib import resources


def load_json(path):
    with resources.open_text("biomage_programmatic_interface", path) as raw_file:
        json_obj = json.load(raw_file)
    return json_obj


def is_file_hidden(path):
    return path.split("/")[-1][0] == "."



from os import listdir
from os.path import isfile, join

from biomage_programmatic_interface.sample_file import SampleFile
from biomage_programmatic_interface.utils import is_file_hidden


class Sample:
    def __init__(self, name):
        self.__name = name
        self.__uuid = None
        self.__sample_files = []

    @property
    def name(self):
        return self.__name

    @property
    def uuid(self):
        return self.__uuid

    @uuid.setter
    def uuid(self, uuid):
        if self.__uuid is not None:
            raise Exception(f"uuid already set for sample {self.__name}")
        self.__uuid = uuid

    def to_json(self):
        return {"name": self.__name, "sampleTechnology": "10x", "options": {}}

    def get_sample_files(self):
        return self.__sample_files

    def add_sample_file(self, sample_file):
        self.__sample_files.append(sample_file)

    @staticmethod
    def __find_all_files_recursively(path):
        file_paths = listdir(path)
        ret = {}
        for file_path in file_paths:
            full_path = join(path, file_path)

            if is_file_hidden(full_path):
                continue

            if isfile(full_path):
                file = SampleFile(full_path)

                if not file.is_valid_type():
                    continue

                folder_name = file.folder

                if ret.get(folder_name) is None:
                    ret[folder_name] = Sample(folder_name)
                ret[folder_name].add_sample_file(file)
                continue

            ret.update(Sample.__find_all_files_recursively(full_path))
        return ret

    @staticmethod
    def get_all_samples_from_path(path):
        return Sample.__find_all_files_recursively(path).values()



import datetime
import hashlib

from biomage_programmatic_interface.sample import Sample


class Experiment:
    @staticmethod
    def create_experiment(connection, name=None):
        created_at = datetime.datetime.now().isoformat()
        hashed_string = hashlib.md5(created_at.encode())
        id = hashed_string.hexdigest()
        name = name if name else id
        experiment_data = {
            "id": id,
            "name": name,
            "description": "",
        }

        connection.fetch_api("v2/experiments/" + id, body=experiment_data)
        return Experiment(connection, id, name)

    def __init__(self, connection, id, name):
        self.__connection = connection
        self.__id = id
        self.__name = name

    @property
    def id(self):
        return self.__id

    @property
    def name(self):
        return self.__name

    def __notify_upload(self, sample_id, sample_file_type):
        url = f"v2/experiments/{self.id}/samples/{sample_id}/sampleFiles/{sample_file_type}"
        json = {"uploadStatus": "uploaded"}
        self.__connection.fetch_api(url, json, "PATCH")

    def __create_sample_file(self, sample_uuid, sample_file):
        url = f"v2/experiments/{self.id}/samples/{sample_uuid}/sampleFiles/{sample_file.get_type()}"
        response = self.__connection.fetch_api(url, sample_file.to_json())
        return response.content

    def __create_samples(self, samples):
        url = f"v2/experiments/{self.id}/samples"
        body = [sample.to_json() for sample in samples]

        sample_ids_by_name = self.__connection.fetch_api(url, body).json()

        for sample in samples:
            sample.uuid = sample_ids_by_name[sample.name]

    def __upload_sample(self, sample):
        url = f"v2/experiments/{self.id}/samples/{sample.uuid}"
        self.__connection.fetch_api(url, sample.to_json())

        for sample_file in sample.get_sample_files():
            s3url_raw = self.__create_sample_file(
                sample.uuid,
                sample_file,
            )
            s3url = s3url_raw.decode("utf-8").replace('"', "")
            self.__connection.uploadS3(sample_file, s3url)

            self.__notify_upload(sample.uuid, sample_file.get_type())

    def upload_samples(self, samples_path):
        samples = Sample.get_all_samples_from_path(samples_path)

        self.__create_samples(samples)

        for sample in samples:
            try:
                self.__upload_sample(sample)
            except Exception as e:
                print(
                    f"Upload failed: {e}. This is likely an error within \
                     the python package for uploading."
                )
                print(
                    "Please send an email to hello@biomage.net and we \
                     will try to resolve this problem as soon as possible"
                )



import uuid
from os.path import getsize

from biomage_programmatic_interface.objectS3 import ObjectS3


class SampleFile(ObjectS3):
    def __init__(self, path):
        super().__init__(path)
        self.__uuid = str(uuid.uuid4())

    @classmethod
    def from_sample_file(cls, sample_file):
        return SampleFile(sample_file.path)

    @property
    def name(self):
        return self.path.split("/")[-1]

    @property
    def uuid(self):
        return self.__uuid

    @property
    def folder(self):
        return self.path.split("/")[-2]

    def get_size(self):
        return getsize(self.path)

    def is_valid_type(self):
        return self.get_type() is not None

    def get_type(self):
        file_types = {
            "matrix.mtx": "matrix10x",
            "barcodes.tsv": "barcodes10x",
            "features.tsv": "features10x",
            "genes.tsv": "features10x",
        }

        for file_type_key in file_types.keys():
            if file_type_key in self.name:
                return file_types[file_type_key]
        return None

    def to_json(self):
        return {
            "sampleFileId": self.__uuid,
            "size": self.get_size(),
        }



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()

setup(
    name="biomage_programmatic_interface",
    version='0.0.27',
    author="Biomage Ltd",
    author_email="engineering@biomage.net",
    description="For programmatic upload of files to Cellencis",
    url = "https://github.com/biomage-org/programmatic_interface",
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    keywords=['pypi', 'cicd', 'python'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License"
    ]
)


import backoff
import boto3
import requests

import biomage_programmatic_interface.exceptions as exceptions
from biomage_programmatic_interface.experiment import Experiment

HTTP_METHODS = {"POST": requests.post, "PATCH": requests.patch, "GET": requests.get}


class Connection:
    def __init__(self, username, password, instance_url, verbose=True):
        self.verbose = verbose
        self.__api_url = self.__get_api_url(instance_url)
        self.username = username
        self.password = password
        self.instance_url = instance_url
        cognito_params = self.__get_cognito_params().json()
        self.clientId = cognito_params["clientId"]
        self.region = cognito_params["clientRegion"]
        self.__authenticate()

    def __get_cognito_params(self):
        try:
            return requests.get(self.__api_url + "v2/programmaticInterfaceClient")
        except Exception:
            raise exceptions.InstanceNotFound() from None

    def __authenticate(self):
        client = boto3.client("cognito-idp", region_name=self.region)

        try:
            resp = client.initiate_auth(
                ClientId=self.clientId,
                AuthFlow="USER_PASSWORD_AUTH",
                AuthParameters={"USERNAME": self.username, "PASSWORD": self.password},
            )
        except Exception:
            raise exceptions.IncorrectCredentials() from None

        print("Authentication succesfull") if self.verbose else ""

        self.__jwt = resp["AuthenticationResult"]["IdToken"]

    def __get_api_url(self, instance_url):
        if instance_url == "local":
            return "http://localhost:3000/"
        if instance_url.startswith("https://"):
            return instance_url
        return f"https://api.{instance_url}/"

    @backoff.on_exception(
        backoff.constant,
        requests.exceptions.HTTPError,
        max_tries=3,
        jitter=backoff.full_jitter,
    )
    def fetch_api(self, url, body, method="POST"):
        headers = {
            "Authorization": "Bearer " + self.__jwt,
            "Content-Type": "application/json",
        }

        response = HTTP_METHODS[method](
            self.__api_url + url, json=body, headers=headers
        )

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if response.status_code == 401:
                print("fetch_api: refresh expired token")
                self.__authenticate()
            raise e

        return response

    def uploadS3(self, objectS3, signed_url, compress=True):
        if compress and not objectS3.is_compressed():
            objectS3.compress()
        headers = {"Content-type": "application/octet-stream"}
        with open(objectS3.path, "rb") as file:
            requests.put(signed_url, headers=headers, data=file.read())

        print(f"Uploaded {objectS3.path} to S3") if self.verbose else ""

    def create_experiment(self, experiment_name=None):
        experiment = Experiment.create_experiment(self, experiment_name)
        print(f"Experiment {experiment.name} created!") if self.verbose else ""
        return experiment
